import dataclasses
import datetime
import typing


@dataclasses.dataclass
class VideoHls:
    url: typing.Optional[str] = None
    info: typing.Optional[str] = None
    info_code: typing.Optional[int] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'VideoHls':
        return cls(info_code=data.pop('infoCode'), **data)


@dataclasses.dataclass
class Owner:
    id: typing.Optional[int] = None
    nickname: typing.Optional[str] = None
    avatar_url: typing.Optional[str] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Owner':
        cls(
            avatar_url=data.pop('avatarUrl'),
            **data,
        )


@dataclasses.dataclass
class Likes:
    sum: typing.Optional[int] = None
    negative: typing.Optional[int] = None
    positive: typing.Optional[int] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Likes':
        return cls(**data)


@dataclasses.dataclass
class Comment:
    id: typing.Optional[str] = None
    created_on: typing.Optional[int] = None
    body: typing.Optional[str] = None
    owner: typing.Optional[Owner] = None
    likes: typing.Optional[Likes] = None
    replies: typing.Optional[typing.List['Comment']] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Comment':
        return cls(
            created_on=data.pop('createdOn'),
            replies=[Comment.from_dict(reply) for reply in data.pop('replies', [])],
            owner=Owner.from_dict(data.pop('owner', {})),
            likes=Likes.from_dict(data.pop('likes', {})),
            **data,
        )


@dataclasses.dataclass
class BodyItems:
    type: typing.Optional[str] = None
    body: typing.Optional[str] = None
    id: typing.Optional[int] = None
    index: typing.Optional[int] = None
    data: typing.Optional[str] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'BodyItems':
        return cls(**data)


@dataclasses.dataclass
class Site:
    id: typing.Optional[int] = None
    name: typing.Optional[str] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Site':
        return cls(**data)


@dataclasses.dataclass
class Section:
    id: typing.Optional[int] = None
    title: typing.Optional[str] = None
    site_id: typing.Optional[int] = None
    meta_keywords: typing.Optional[str] = None
    meta_description: typing.Optional[str] = None
    window_title: typing.Optional[str] = None
    gemius_code: typing.Optional[datetime.datetime] = None
    path: typing.Optional[str] = None
    root_section_id: typing.Optional[int] = None
    grid_section_id: typing.Optional[int] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Section':
        return cls(
            id=data.get('id'),
            title=data.get('title'),
            site_id=data.get('siteId'),
            meta_keywords=data.get('metaKeywords'),
            meta_description=data.get('metaDescription'),
            window_title=data.get('windowTitle'),
            gemius_code=data.get('gemiusCode'),
            path=data.get('path'),
            root_section_id=data.get('rootSectionId'),
            grid_section_id=data.get('gridSectionId'),
        )


@dataclasses.dataclass
class FrontImage:
    id: typing.Optional[int] = None
    src: typing.Optional[str] = None
    caption: typing.Optional[str] = None
    source: typing.Optional[str] = None
    type: typing.Optional[str] = None
    order: typing.Optional[int] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'FrontImage':
        return cls(**data)


@dataclasses.dataclass
class Images:
    caption: typing.Optional[str] = None
    show_caption: typing.Optional[bool] = None
    source: typing.Optional[str] = None
    src: typing.Optional[str] = None
    type: typing.Optional[str] = None
    link: typing.Optional[str] = None
    id: typing.Optional[int] = None
    height: typing.Optional[int] = None
    width: typing.Optional[int] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Images':
        return cls(
            show_caption=data.pop('showCaption'),
            **data,
        )


@dataclasses.dataclass
class Image:
    src: typing.Optional[str] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Image':
        return cls(**data)


@dataclasses.dataclass
class Videos:
    id: typing.Optional[int] = None
    object_type: typing.Optional[str] = None
    title: typing.Optional[str] = None
    subtype: typing.Optional[str] = None
    date: typing.Optional[int] = None
    order: typing.Optional[int] = None
    length: typing.Optional[int] = None
    style: typing.Optional[str] = None
    preroll: typing.Optional[bool] = None
    postroll: typing.Optional[bool] = None
    vtt: typing.Optional[str] = None
    image: typing.Optional[Image] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Videos':
        return cls(
            image=Image.from_dict(data.pop('image', {})),
            object_type=data.pop('objectType'),
            **data,
        )


@dataclasses.dataclass
class Galleries:
    id: typing.Optional[int] = None
    title: typing.Optional[str] = None
    summary: typing.Optional[str] = None
    type: typing.Optional[str] = None
    subtype: typing.Optional[str] = None
    images: typing.Optional[typing.List[Images]] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Galleries':
        return cls(images=[Images.from_dict(image) for image in data.pop('images', [])], **data)


@dataclasses.dataclass
class Quotes:
    order: typing.Optional[int] = None
    author: typing.Optional[str] = None
    body: typing.Optional[str] = None
    type: typing.Optional[int] = None
    url: typing.Optional[str] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Quotes':
        return cls(**data)


@dataclasses.dataclass
class Article:
    id: typing.Optional[int] = None
    title: typing.Optional[str] = None
    subtitle: typing.Optional[str] = None
    summary: typing.Optional[str] = None
    keywords: typing.Optional[str] = None
    categories: typing.Optional[str] = None
    place: typing.Optional[str] = None
    date: typing.Optional[int] = None
    url: typing.Optional[str] = None
    nb_views: typing.Optional[int] = None
    last_changed: typing.Optional[int] = None
    style: typing.Optional[str] = None
    desktop_url: typing.Optional[str] = None
    has_banners: typing.Optional[int] = None
    has_comments: typing.Optional[int] = None
    allow_new_comments: typing.Optional[int] = None
    has_votes: typing.Optional[str] = None
    is_slim: typing.Optional[bool] = None
    body_items: typing.Optional[typing.List[BodyItems]] = None
    has_videos: typing.Optional[int] = None
    has_galleries: typing.Optional[int] = None
    has_streams: typing.Optional[int] = None
    is_developing: typing.Optional[bool] = None
    has_audios: typing.Optional[int] = None
    has_documents: typing.Optional[int] = None
    nb_comments: typing.Optional[int] = None
    has_active_feed: typing.Optional[bool] = None
    substyle: typing.Optional[str] = None
    reading_time: typing.Optional[int] = None
    source: typing.Optional[str] = None
    feeds: typing.Optional[typing.List] = None
    front_image: typing.Optional[FrontImage] = None
    additional_front_images: typing.List[FrontImage] = None
    images: typing.Optional[typing.List[Images]] = None
    videos: typing.Optional[typing.List[Videos]] = None
    audios: typing.Optional[typing.List] = None
    streams: typing.Optional[typing.List] = None
    documents: typing.Optional[typing.List] = None
    galleries: typing.Optional[typing.List[Galleries]] = None
    section: typing.Optional[Section] = None
    quotes: typing.Optional[typing.List[Quotes]] = None
    embeds: typing.Optional[typing.List] = None
    polls: typing.Optional[typing.List] = None
    site: typing.Optional[Site] = None
    related_articles: typing.Optional[typing.List] = None
    related_recipes: typing.Optional[typing.List] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Article':
        return cls(
            id=data.get('id'),
            title=data.get('title'),
            subtitle=data.get('subtitle'),
            summary=data.get('summary'),
            keywords=data.get('keywords'),
            categories=data.get('categories'),
            place=data.get('place'),
            date=data.get('date'),
            url=data.get('url'),
            nb_views=data.get('nbViews'),
            last_changed=data.get('lastChanged'),
            style=data.get('style'),
            desktop_url=data.get('desktopUrl'),
            has_banners=data.get('hasBanners'),
            has_comments=data.get('hasComments'),
            allow_new_comments=data.get('allowNewComments'),
            has_votes=data.get('hasVotes'),
            is_slim=data.get('isSlim'),
            body_items=[BodyItems.from_dict(body_item) for body_item in data.get('bodyItems', [])],
            has_videos=data.get('hasVideos'),
            has_galleries=data.get('hasGalleries'),
            has_streams=data.get('hasStreams'),
            is_developing=data.get('isDeveloping'),
            has_audios=data.get('hasAudios'),
            has_documents=data.get('hasDocuments'),
            nb_comments=data.get('nbComments'),
            has_active_feed=data.get('hasActiveFeed'),
            substyle=data.get('substyle'),
            reading_time=data.get('readingTime'),
            source=data.get('source'),
            feeds=data.get('feeds'),
            front_image=FrontImage.from_dict(data.get('frontImage', {})),
            additional_front_images=[FrontImage(**image) for image in data.get('additionalFrontImages', [])],
            images=[Images.from_dict(image) for image in data.get('images', [])],
            videos=[Videos.from_dict(video) for video in data.get('videos', [])],
            audios=data.get('audios'),
            streams=data.get('streams'),
            documents=data.get('documents'),
            galleries=[Galleries.from_dict(gallery) for gallery in data.get('galleries', [])],
            section=Section.from_dict(data.get('section', {})),
            quotes=[Quotes.from_dict(quote) for quote in data.get('quotes', [])],
            embeds=data.get('embeds'),
            polls=data.get('polls'),
            site=Site.from_dict(data.get('site', {})),
            related_articles=data.get('relatedArticles'),
            related_recipes=data.get('relatedRecipes'),
        )


@dataclasses.dataclass
class Data:
    article: typing.Optional[Article] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Data':
        return cls(article=Article.from_dict(data.get('article', {})))


@dataclasses.dataclass
class Root:
    data: typing.Optional[Data] = None

    @classmethod
    def from_dict(cls, data: typing.Dict) -> 'Root':
        return cls(data=Data.from_dict(data.get('data', {})))



import dataclasses
import datetime
import typing


@dataclasses.dataclass
class Image:
    caption: typing.Optional[str] = None
    url: typing.Optional[str] = None
    height: typing.Optional[int] = None
    width: typing.Optional[int] = None


@dataclasses.dataclass
class Video:
    title: typing.Optional[str] = None
    url: typing.Optional[str] = None


@dataclasses.dataclass
class Comment:
    posted_at: typing.Optional[datetime.datetime] = None
    content: typing.Optional[str] = None
    author: typing.Optional[str] = None
    author_avatar_url: typing.Optional[str] = None
    likes: typing.Optional[int] = None
    dislikes: typing.Optional[int] = None
    score: typing.Optional[int] = None
    replies: typing.Optional[typing.List['Comment']] = None


@dataclasses.dataclass
class Article:
    id: typing.Optional[str] = None
    title: typing.Optional[str] = None
    summary: typing.Optional[str] = None
    content: typing.Optional[str] = None
    place: typing.Optional[str] = None
    num_views: typing.Optional[int] = None
    images: typing.Optional[typing.List[Image]] = None
    videos: typing.Optional[typing.List[Video]] = None
    posted_at: typing.Optional[datetime.datetime] = None
    comments: typing.Optional[typing.List[Comment]] = None



import datetime
import io
import sys
import typing
import uuid

import aiohttp

from api_24ur import downloader
from api_24ur import schemas
from api_24ur import types


ARTICLE_API_URL = 'https://gqlc.24ur.com/graphql/?raw=&query=onl_web_full_article(url:%20%22{path}%22)'
ARTICLE_HEADERS = {
    'host': 'gqlc.24ur.com',
    'accept-encoding': 'gzip',
}

GRAPH_API_URL = 'https://gql.24ur.si/graphql'
GRAPH_API_HEADERS = {
    'accept-encoding': 'gzip',
    'content-type': 'application/graphql; charset=utf-8',
    'host': 'gql.24ur.si',
}
VIDEO_STREAM_PAYLOAD = '{{videoHlsUrl(id: {video_id}, siteId: 1) {{url info infoCode}}}}'
COMMENTS_PAYLOAD = '{{ comments(itemType: ARTICLE, itemId: {article_id}, perPage: {limit}, nextId: "") {{ total totalShown nextId comments {{ id createdOn body owner {{ id nickname avatarUrl }} likes {{ sum negative positive }} replies {{ id body createdOn owner {{ id nickname avatarUrl }} likes {{ sum negative positive }}}}}}}}}}'  # noqa: E501

DEFAULT_USER_AGENT = '24ur Android App 4.4.1 (168)'


class Client:
    def __init__(
        self,
        user_agent: typing.Optional[str] = None,
        device_id: typing.Optional[str] = None,
        authorization: typing.Optional[str] = None,
    ) -> None:
        self._article_headers = ARTICLE_HEADERS | {
            'device-id': device_id or str(uuid.uuid4()),
            'authorization': authorization or '',
            'user-agent': user_agent or DEFAULT_USER_AGENT,
        }
        self._graph_headers = GRAPH_API_HEADERS | self._article_headers

    async def get_article_by_url(
        self,
        url: str,
        num_comments: int = 0,
        image_size_px: int = 1200,
    ) -> types.Article:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=ARTICLE_API_URL.format(path=self._path_from_url(url)),
                headers=self._article_headers,
            ) as response:
                data = await response.json()

        article = schemas.Root.from_dict(data).data.article

        return types.Article(
            id=article.id,
            title=article.title,
            summary=article.summary,
            content='\n'.join([item.body if item.body else '' for item in article.body_items or []]),
            place=article.place,
            num_views=article.nb_views,
            images=[
                types.Image(
                    caption=image.caption,
                    url=image.src.replace('PLACEHOLDER', f'{image_size_px}xX'),
                    height=image.height,
                    width=image.width,
                )
                for image in article.images or []
            ],
            videos=[
                types.Video(
                    title=video.title,
                    url=await self._fetch_video_stream_url(video.id),
                )
                for video in article.videos
            ],
            comments=await self._fetch_comments(article_id=article.id, limit=num_comments) if num_comments > 0 else [],
        )

    async def download_video(
        self,
        stream_url: str,
        download_path: str = '/tmp',
        tmp_dir: str = '/tmp',
        pool_size: int = 5,
        max_bitrate: int = sys.maxsize,
    ) -> str:
        await downloader.Downloader(
            url=stream_url,
            download_path=download_path,
            tmp_dir=tmp_dir,
            pool_size=pool_size,
            max_bitrate=max_bitrate,
        ).download()

    async def download_video_bytes(
        self,
        stream_url: str,
        tmp_dir: str = '/tmp',
        pool_size: int = 5,
        max_bitrate: int = sys.maxsize,
    ) -> io.BytesIO:
        return await downloader.Downloader(
            url=stream_url,
            download_path=tmp_dir,
            tmp_dir=tmp_dir,
            pool_size=pool_size,
            max_bitrate=max_bitrate,
        ).download_bytes()

    async def _fetch_video_stream_url(self, video_id: int) -> str:
        request_body = VIDEO_STREAM_PAYLOAD.format(video_id=video_id)
        headers = self._graph_headers | {'content-length': str(len(request_body))}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=GRAPH_API_URL,
                data=request_body,
                headers=headers,
            ) as response:
                data = await response.json()

        return schemas.VideoHls.from_dict(data.get('data', {}).get('videoHlsUrl', {})).url

    async def _fetch_comments(self, article_id: int, limit: int = 1000) -> typing.List[types.Comment]:
        request_body = COMMENTS_PAYLOAD.format(article_id=article_id, limit=limit)
        headers = self._graph_headers | {'content-length': str(len(request_body))}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=GRAPH_API_URL,
                data=request_body,
                headers=headers,
            ) as response:
                data = await response.json()

        return self._parse_comments(
            [
                schemas.Comment.from_dict(comment)
                for comment in data.get('data', {}).get('comments', {}).get('comments', [])
            ]
        )

    def _path_from_url(self, url: str) -> str:
        return url.split('?')[0].split('.com')[1]

    def _parse_comments(self, comments: typing.List[schemas.Comment]) -> typing.List[types.Comment]:
        return [
            types.Comment(
                author=comment.owner.nickname if comment.owner else 'Unknown',
                author_avatar_url=comment.owner.avatar_url if comment.owner else None,
                posted_at=datetime.datetime.fromtimestamp(comment.created_on) if comment.created_on else None,
                content=comment.body,
                likes=comment.likes.positive,
                dislikes=comment.likes.negative,
                score=comment.likes.sum,
                replies=self._parse_comments(comment.replies) if comment.replies else [],
            )
            for comment in comments
        ]



import io
import logging
import os
import shutil
import sys
import uuid

import aiohttp

try:
    from m3u8downloader import main as m3u8
except ImportError as e:
    raise RuntimeError('Download dependencies not installed, please install them') from e

from api_24ur import exceptions


# Monkeypath lib
def get_local_file_for_url(tempdir, url, path_line=None):
    if path_line and path_line.startswith(tempdir):
        return path_line
    path = m3u8.get_url_path(url)
    if path.startswith("/"):
        path = path[1:]

    # Limit file name length
    ext = path.split('.')[-1]
    path = f'{path[:250]}.{ext}'

    return os.path.normpath(os.path.join(tempdir, path))


m3u8.get_local_file_for_url = get_local_file_for_url
m3u8.logger.handlers = [logging.NullHandler()]
m3u8.logger.propagate = False


class Downloader:
    def __init__(
        self,
        url: str,
        download_path: str = '/tmp',
        tmp_dir: str = '/tmp',
        pool_size: int = 5,
        max_bitrate: int = sys.maxsize,
    ) -> None:
        self.url = url
        self.download_path = download_path
        self.tmp_dir = tmp_dir
        self.pool_size = pool_size
        self.max_bitrate = max_bitrate

    async def download(self) -> str:
        output_file_path = f'{self.download_path}/{uuid.uuid4()}.mp4'

        downloader = m3u8.M3u8Downloader(
            url=await self._find_suitable_stream(),
            output_filename=output_file_path,
            tempdir=self.tmp_dir,
        )
        downloader.start()

        # Cleanup
        shutil.rmtree(path=downloader.tempdir, ignore_errors=True)

        return output_file_path

    async def download_bytes(self) -> io.BytesIO:
        fp = await self.download()
        with open(fp, 'rb') as f:
            bytes = io.BytesIO(f.read())

        os.remove(fp)
        return bytes

    async def _find_suitable_stream(self) -> str:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=self.url,
            ) as response:
                data = (await response.text()).split('\n')

        stream, bitrate = None, 0
        for i in range(0, len(data), 2):
            for part in data[i].split(','):
                key, _, val = part.partition('=')
                if key == 'BANDWIDTH' and int(val) > bitrate and int(val) <= self.max_bitrate:
                    bitrate = int(val)
                    stream = data[i + 1]

        if not stream:
            raise exceptions.DownloadException('No suitable stream found')

        return stream
